{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 14](https://adventofcode.com/2019/day/14) puzzle input is a set of sequences (mine is [here](puzzle_input/day14.txt)) like \"`10 ORE => 10A`\", which would mean \"You can produce 10 of resource 'A' using 10 of the'ORE' resource\". Part 1 involves using these reactions to determine the minimum amount of ORE required to produce 1 \"FUEL\". Part 2 involves going the other way - finding how much FUEL you can produce with 1 trillion ORE."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "374457\n"
     ]
    }
   ],
   "source": [
    "quantity_idx = 0\n",
    "component_idx = 1\n",
    "\n",
    "class ReactionResult:\n",
    "    def __init__(self, quantity_produced, required):\n",
    "        self.quantity_produced = quantity_produced\n",
    "        self.required = required\n",
    "        \n",
    "    def __repr__(self):\n",
    "        if not self.required:\n",
    "            return str(self.quantity_produced)\n",
    "        return f\"{self.quantity_produced} <= (\" + \", \".join([f\"{qty}x{comp}\" for qty,comp in self.required]) + \")\"\n",
    "        \n",
    "def parse_component(component_str):\n",
    "    quantity_str, component_str = component_str.strip().split(\" \")\n",
    "    return (int(quantity_str), component_str)\n",
    "\n",
    "def parse_reactions(reactions_list_str):\n",
    "    reactions = {\n",
    "        \"ORE\": ReactionResult(1, None)\n",
    "    }\n",
    "    \n",
    "    for reaction_str in reactions_list_str:\n",
    "        components_str, result_str = reaction_str.split(\"=>\")\n",
    "        result_amount, result_component = parse_component(result_str)\n",
    "        components = [ parse_component(component_str) for component_str in components_str.split(\",\") ]\n",
    "        reactions[result_component] = ReactionResult(result_amount, components)\n",
    "        \n",
    "    return reactions\n",
    "\n",
    "\n",
    "def ore_for_component(reactions, quantity, component, spares, level=0, debug=False):\n",
    "    if component in spares:\n",
    "        # if the amount required or more is spare we don't need any more reactions\n",
    "        # so just update the structure and return 0\n",
    "        if spares[component] >= quantity:\n",
    "            spares[component] -= quantity\n",
    "            return 0\n",
    "        \n",
    "        # if we have less spare, than is required, then just subtract the \n",
    "        # spare amount from the quantity required and continue with the lesser\n",
    "        # quantity\n",
    "        elif spares[component] < quantity:\n",
    "            quantity -= spares[component]\n",
    "            spares[component] = 0\n",
    "            \n",
    "    if component not in reactions:\n",
    "        raise Exception(f\"Unknown component {component}\")\n",
    "        \n",
    "    reaction = reactions[component]\n",
    "\n",
    "    reaction_count = int(quantity / reaction.quantity_produced)\n",
    "    if (reaction_count * reaction.quantity_produced) < quantity:\n",
    "        reaction_count += 1\n",
    "        \n",
    "    # if we'll produce more than we need, add some to the spares registry\n",
    "    if reaction_count * reaction.quantity_produced > quantity:\n",
    "        spare_quantity = (reaction_count * reaction.quantity_produced) - quantity\n",
    "        if component in spares:\n",
    "            spares[component] += spare_quantity\n",
    "        else:\n",
    "            spares[component] = spare_quantity\n",
    "\n",
    "    if debug:\n",
    "        left_padding = \"\".join(level * [\"\\t\"])    \n",
    "        print(f\"{left_padding}{quantity} x {component}\")\n",
    "        \n",
    "    # we have a raw material which requires no reactions, so just return\n",
    "    # the quantity that was requested \n",
    "    if not reaction.required:\n",
    "        return quantity\n",
    "    \n",
    "    \n",
    "    total_ore = 0\n",
    "    \n",
    "    for required_quantity, required_component in reaction.required:\n",
    "        total_ore += ore_for_component(reactions, reaction_count * required_quantity, required_component, spares, 1 + level, debug)\n",
    "    \n",
    "    return total_ore\n",
    "    \n",
    "\n",
    "test_reactions0 = [\n",
    "    \"10 ORE => 10 A\",\n",
    "    \"1 ORE => 1 B\",\n",
    "    \"7 A, 1 B => 1 C\",\n",
    "    \"7 A, 1 C => 1 D\",\n",
    "    \"7 A, 1 D => 1 E\",\n",
    "    \"7 A, 1 E => 1 FUEL\"\n",
    "]\n",
    "                     \n",
    "test_reactions1 = [\n",
    "    \"9 ORE => 2 A\",\n",
    "    \"8 ORE => 3 B\",\n",
    "    \"7 ORE => 5 C\",\n",
    "    \"3 A, 4 B => 1 AB\",\n",
    "    \"5 B, 7 C => 1 BC\",\n",
    "    \"4 C, 1 A => 1 CA\",\n",
    "    \"2 AB, 3 BC, 4 CA => 1 FUEL\"\n",
    "]\n",
    "\n",
    "test_reactions2 = [\n",
    "    \"157 ORE => 5 NZVS\",\n",
    "    \"165 ORE => 6 DCFZ\",\n",
    "    \"44 XJWVT, 5 KHKGT, 1 QDVJ, 29 NZVS, 9 GPVTF, 48 HKGWZ => 1 FUEL\",\n",
    "    \"12 HKGWZ, 1 GPVTF, 8 PSHF => 9 QDVJ\",\n",
    "    \"179 ORE => 7 PSHF\",\n",
    "    \"177 ORE => 5 HKGWZ\",\n",
    "    \"7 DCFZ, 7 PSHF => 2 XJWVT\",\n",
    "    \"165 ORE => 2 GPVTF\",\n",
    "    \"3 DCFZ, 7 NZVS, 5 HKGWZ, 10 PSHF => 8 KHKGT\"\n",
    "]\n",
    "\n",
    "test_reactions3 = [\n",
    "    \"2 VPVL, 7 FWMGM, 2 CXFTF, 11 MNCFX => 1 STKFG\",\n",
    "    \"17 NVRVD, 3 JNWZP => 8 VPVL\",\n",
    "    \"53 STKFG, 6 MNCFX, 46 VJHF, 81 HVMC, 68 CXFTF, 25 GNMV => 1 FUEL\",\n",
    "    \"22 VJHF, 37 MNCFX => 5 FWMGM\",\n",
    "    \"139 ORE => 4 NVRVD\",\n",
    "    \"144 ORE => 7 JNWZP\",\n",
    "    \"5 MNCFX, 7 RFSQX, 2 FWMGM, 2 VPVL, 19 CXFTF => 3 HVMC\",\n",
    "    \"5 VJHF, 7 MNCFX, 9 VPVL, 37 CXFTF => 6 GNMV\",\n",
    "    \"145 ORE => 6 MNCFX\",\n",
    "    \"1 NVRVD => 8 CXFTF\",\n",
    "    \"1 VJHF, 6 MNCFX => 4 RFSQX\",\n",
    "    \"176 ORE => 6 VJHF\",\n",
    "]\n",
    "\n",
    "test_reactions4 = [\n",
    "    \"171 ORE => 8 CNZTR\",\n",
    "    \"7 ZLQW, 3 BMBT, 9 XCVML, 26 XMNCP, 1 WPTQ, 2 MZWV, 1 RJRHP => 4 PLWSL\",\n",
    "    \"114 ORE => 4 BHXH\",\n",
    "    \"14 VRPVC => 6 BMBT\",\n",
    "    \"6 BHXH, 18 KTJDG, 12 WPTQ, 7 PLWSL, 31 FHTLT, 37 ZDVW => 1 FUEL\",\n",
    "    \"6 WPTQ, 2 BMBT, 8 ZLQW, 18 KTJDG, 1 XMNCP, 6 MZWV, 1 RJRHP => 6 FHTLT\",\n",
    "    \"15 XDBXC, 2 LTCX, 1 VRPVC => 6 ZLQW\",\n",
    "    \"13 WPTQ, 10 LTCX, 3 RJRHP, 14 XMNCP, 2 MZWV, 1 ZLQW => 1 ZDVW\",\n",
    "    \"5 BMBT => 4 WPTQ\",\n",
    "    \"189 ORE => 9 KTJDG\",\n",
    "    \"1 MZWV, 17 XDBXC, 3 XCVML => 2 XMNCP\",\n",
    "    \"12 VRPVC, 27 CNZTR => 2 XDBXC\",\n",
    "    \"15 KTJDG, 12 BHXH => 5 XCVML\",\n",
    "    \"3 BHXH, 2 VRPVC => 7 MZWV\",\n",
    "    \"121 ORE => 7 VRPVC\",\n",
    "    \"7 XCVML => 6 RJRHP\",\n",
    "    \"5 BHXH, 4 VRPVC => 5 LTCX\"\n",
    "]\n",
    "\n",
    "assert 31 == ore_for_component(parse_reactions(test_reactions0), 1, \"FUEL\", {})\n",
    "assert 165 == ore_for_component(parse_reactions(test_reactions1), 1, \"FUEL\", {})\n",
    "assert 13312 == ore_for_component(parse_reactions(test_reactions2), 1, \"FUEL\", {})\n",
    "assert 180697 == ore_for_component(parse_reactions(test_reactions3), 1, \"FUEL\", {})\n",
    "assert 2210736 == ore_for_component(parse_reactions(test_reactions4), 1, \"FUEL\", {})\n",
    "\n",
    "puzzle_reactions = [ l.strip() for l in open(\"puzzle_input/day14.txt\", \"r\").readlines() ]\n",
    "print(ore_for_component(parse_reactions(puzzle_reactions), 1, \"FUEL\", {}))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3568888\n"
     ]
    }
   ],
   "source": [
    "one_trillion = 1_000_000_000_000\n",
    "\n",
    "# yes we are going to brute force this one too, thankyouverymuch\n",
    "def find_a_trillion_ore(increment, reactions_table, ore_target):\n",
    "    \n",
    "    current_guess = increment   \n",
    "    ore_required = ore_for_component(parse_reactions(reactions_table), current_guess, \"FUEL\", {}, 0, False)\n",
    "    \n",
    "    previous_guess = one_trillion - ore_required\n",
    "    \n",
    "    while ore_required != ore_target:\n",
    "        if ore_required > ore_target:\n",
    "            # if we overshot it then go back, step down the increment a bit and try again\n",
    "            current_guess -= increment\n",
    "            increment = int(increment / 2)\n",
    "            \n",
    "        else:\n",
    "            # it's late and this is 99/100 gonna be the answer so it'll have to do\n",
    "            if previous_guess == current_guess and increment == 1:\n",
    "                break\n",
    "            \n",
    "            current_guess += increment\n",
    "            \n",
    "        \n",
    "        ore_required = ore_for_component(parse_reactions(reactions_table), current_guess, \"FUEL\", {}, 0, False)\n",
    "        difference = one_trillion - ore_required\n",
    "        previous_guess = current_guess\n",
    "        \n",
    "    return current_guess\n",
    "\n",
    "# assert 82892753 == find_a_trillion_ore(65536, test_reactions2, one_trillion)\n",
    "assert 5586022 == find_a_trillion_ore(65536, test_reactions3, one_trillion)\n",
    "assert 460664 == find_a_trillion_ore(65536, test_reactions4, one_trillion)\n",
    "\n",
    "print(find_a_trillion_ore(65536, puzzle_reactions, one_trillion))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
